# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Initilization Instance() with VM information. """

from googleapiclient.discovery import Resource

from dataclasses import dataclass, field
from typing import Dict, List, Union
from time import time
from gce_rescue.tasks.backup import backup_metadata_items
from gce_rescue.tasks.disks import list_disk
from gce_rescue.tasks.pre_validations import Validations
from gce_rescue.config import get_config


def get_instance_info(
  compute: Resource,
  name: str,
  project_data: Dict[str, str]
) -> Dict:
  """Set Dictionary with complete data from instances().get() from the instance.
  https://cloud.google.com/compute/docs/reference/rest/v1/instances/get
  Attributes:
    compute: obj, API Object
    instance: str, Instace name
    project_data: dict, Dictionary containing project and zone keys to be
      unpacked when calling the API.
  """
  return compute.instances().get(
      **project_data,
      instance = name).execute()


def guess_guest(data: Dict) -> str:
  """Determined which Guest OS Family is being used and select a
  different OS for recovery disk.
     Default: projects/debian-cloud/global/images/family/debian-11"""

  guests = get_config('source_guests')
  for disk in data['disks']:
    if disk['boot']:
      if 'architecture' in disk:
        arch = disk['architecture'].lower()
      else:
        arch = 'x86_64'
      guest_default = guests[arch][0]
      guest_name = guest_default.split('/')[-1]
      for lic in disk['licenses']:
        if guest_name in lic:
          guest_default = guests[arch][1]
  return guest_default


def validate_instance_mode(data: Dict) -> Dict:
  """Validate if the instance is already configured as rescue mode."""

  result = {
      'rescue-mode': False,
      'ts': generate_ts()
  }
  if 'metadata' in data and  'items' in data['metadata']:
    metadata = data['metadata']
    for item in metadata['items']:
      if item['key'] == 'rescue-mode':
        result = {
          'rescue-mode': True,
          'ts': item['value']
        }

  return result

def generate_ts() -> int:
  """Get the current timestamp to be used as unique ID
  during this execution."""
  return int(time())


@dataclass
class Instance(Resource):
  """Initialize instance."""
  zone: str
  name: str
  project: str = None
  test_mode: bool = field(default_factory=False)
  compute: Resource = field(init=False)
  data: Dict[str, Union[str, int]] = field(init=False)
  ts: int = field(init=False)
  _status: str = ''
  _rescue_source_disk: str = ''
  _rescue_mode_status: Dict[str, Union[str, int]] = field(
    default_factory=lambda: ({})
  )
  _disks: Dict[str, str] = field(default_factory=lambda: ({}))
  _backup_items: Dict[str, Union[str, int]] = field(
    default_factory=lambda: ([])
  )

  def __post_init__(self):
    check = Validations(
        name=self.name,
        test_mode=self.test_mode,
        **self.project_data
    )
    self.compute = check.compute
    self.project = check.adc_project
    self.data = get_instance_info(
        compute=self.compute,
        name=self.name,
        project_data=self.project_data)

    self._rescue_mode_status = validate_instance_mode(self.data)
    self.ts = self._rescue_mode_status['ts']
    self._status = self.data['status']
    self._rescue_source_disk = guess_guest(self.data)
    self._disks = self._define_disks()

    # Backup metadata items
    self._backup_items = backup_metadata_items(
        data=self.data
    )

  def refresh_fingerprint(self) -> None:
    """Refresh the current metadata fingerprint value."""

    project_data = get_instance_info(
        compute=self.compute,
        name=self.name,
        project_data=self.project_data)

    new_fingerprint = project_data['metadata']['fingerprint']
    self.data['metadata']['fingerprint'] = new_fingerprint

  def _define_disks(self) -> Dict[str, str]:
    """Define the values of disk_name and device_name."""

    rescue_on = self._rescue_mode_status['rescue-mode']
    if not rescue_on:
      for disk in self.data['disks']:
        if disk['boot']:
          device_name = disk['deviceName']
          source = disk['source']
          disk_name = source.split('/')[-1]

    else:
      ts = self._rescue_mode_status['ts']
      disk_filter = f'labels.rescue={ts}'

      disk = list_disk(
        vm=self,
        project_data=self.project_data,
        label_filter=disk_filter
      )

      disk_name = disk[0]['name']
      disks = self.data['disks']
      for disk in disks:
        full_source = disk['source']
        source = full_source.split('/')[-1]
        if disk_name == source:
          device_name = disk['deviceName']

    result = {
        'device_name': device_name,
        'disk_name': disk_name
    }
    return result

  @property
  def rescue_mode_status(self) -> Dict[str, Union[str, int]]:
    return self._rescue_mode_status

  @property
  def project_data(self) -> str:
    return {'project': self.project, 'zone': self.zone}

  @property
  def rescue_disk(self) -> str:
    return f'linux-rescue-disk-{self.ts}'

  @property
  def status(self) -> str:
    return self._status

  @status.setter
  def status(self, v: str) -> None:
    self._status = v

  @property
  def rescue_source_disk(self) -> str:
    return self._rescue_source_disk

  @rescue_source_disk.setter
  def rescue_source_disk(self, v: str) -> None:
    self._rescue_source_disk = v

  @property
  def backup_items(self) -> List[str]:
    return self._backup_items

  @backup_items.setter
  def backup_items(self, v: List[str]) -> None:
    self._backup_items = v

  @property
  def disks(self) -> List[str]:
    return self._disks



#!/usr/bin/env python3

# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Main script to be used to set/reset rescue mode. """

from datetime import datetime
import logging

from gce_rescue.config import process_args, set_configs
from gce_rescue import messages
from gce_rescue.gce import Instance
from gce_rescue.tasks.actions import call_tasks
from gce_rescue.utils import read_input, set_logging

def main():
  """ Main script function. """
  parser = process_args()
  args = parser.parse_args()
  set_configs(args)

  set_logging(vm_name=args.name)

  parse_kwargs = {
      'zone': args.zone,
      'name': args.name,
  }

  if args.project:
    parse_kwargs['project'] = args.project

  vm = Instance(test_mode=False, **parse_kwargs)
  rescue_on = vm.rescue_mode_status['rescue-mode']
  if not rescue_on:
    if not args.force:
      info = (f'This option will boot the instance {vm.name} in '
              'RESCUE MODE. \nIf your instance is running it will be rebooted. '
              '\nDo you want to continue [y/N]: ')
      read_input(msg=info)

    print('Starting...')
    # save in the log file current configuration of the VM as backup.
    logging.info('RESTORE#%s\n', vm.data)
    action = 'set_rescue_mode'
    msg = messages.tip_connect_ssh(vm)

  else:
    rescue_ts = vm.rescue_mode_status['ts']
    rescue_date = datetime.fromtimestamp(int(rescue_ts))

    if not args.force:
      info = (f'The instance \"{vm.name}\" is currently configured '
              f'to boot as rescue mode since {rescue_date}.\nWould you like to'
              ' restore the original configuration ? [y/N]: ')
      read_input(msg=info)

    print('Restoring VM...')
    action = 'reset_rescue_mode'
    msg = messages.tip_restore_disk(vm)

  call_tasks(vm=vm, action=action)
  print(msg)


if __name__ == '__main__':
  main()


# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Different functions to modify VM custom metadata. """

from gce_rescue.config import get_config
from gce_rescue.tasks.keeper import wait_for_operation, wait_for_os_boot
from typing import Dict
import logging

_logger = logging.getLogger(__name__)

def set_metadata(vm) -> Dict:
  """Configure Instance custom metadata.
  https://cloud.google.com/compute/docs/reference/rest/v1/instances/setMetadata
    a. Set rescue-mode=<ts unique id> if disable=False
    b. Delete rescue-mode if disable=True
    c. Replace startup-script with local startup-script.sh content."""

  startup_script_file = get_config('startup-script-file')
  device_name = vm.disks['device_name']
  with open(startup_script_file, encoding='utf-8') as file:
    file_content = file.read()
    file_content = file_content.replace('GOOGLE_DISK_NAME', device_name)
    file_content = file_content.replace('GOOGLE_TS', str(vm.ts))

  metadata_body = {
    'fingerprint': vm.data['metadata']['fingerprint'],
    'items': [{
      'key': 'startup-script',
      'value': file_content
    }]
  }
  _logger.info('Setting custom metadata...')

  operation = vm.compute.instances().setMetadata(
    **vm.project_data,
    instance = vm.name,
    body = metadata_body).execute()

  result = wait_for_operation(vm, oper=operation)
  return result


def restore_metadata_items(vm, remove_rescue_mode: bool = False) -> Dict:
  """Restore original metadata.items after the instance is running again."""

  vm.refresh_fingerprint()

  if not remove_rescue_mode:
    vm.backup_items.append({ 'key': 'rescue-mode', 'value': vm.ts })
  else:
    vm.backup_items.remove({ 'key': 'rescue-mode', 'value': vm.ts })

  metadata_body = {
    'fingerprint': vm.data['metadata']['fingerprint'],
    'items': vm.backup_items
  }
  _logger.info('Restoring original metadata...')
  if not remove_rescue_mode and not wait_for_os_boot(vm):
    raise Exception('Guest OS boot timeout.')
  operation = vm.compute.instances().setMetadata(
    **vm.project_data,
    instance = vm.name,
    body = metadata_body).execute()
  result = wait_for_operation(vm, oper=operation)
  return result



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Multitask test code """

from absl.testing import absltest
from gce_rescue.utils import ThreadHandler as Handler
import time

OUTPUT1 = 'task1 done'
OUTPUT2 = 'task1 done'


class MultitasksTest(absltest.TestCase):
  status = {
	  'task1_done': False,
	  'task2_done': False,
	}


  @classmethod
  def task1(cls):
    for _ in range(0, 3):
      # print(f'task 1, time {i}')
      time.sleep(0.001)
    MultitasksTest.status['task1_done'] = True
    return OUTPUT1


  @classmethod
  def task2(cls):
    for _ in range(0, 3):
      # print(f'task 2, time {i}')
      time.sleep(0.001)
    MultitasksTest.status['task2_done'] = True
    return OUTPUT2


  def test_multitasks(self):
    t1 = Handler(target=MultitasksTest.task1)
    t2 = Handler(target=MultitasksTest.task2)
    t1.start()
    t2.start()
    output1 = t1.result()
    output2 = t2.result()
    self.assertEqual(output1, OUTPUT1)
    self.assertEqual(output2, OUTPUT2)
    self.assertTrue(MultitasksTest.status['task1_done'])
    self.assertTrue(MultitasksTest.status['task2_done'])


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test code for backup.py."""

from absl.testing import absltest
from gce_rescue.tasks import backup
from gce_rescue.test.mocks import mock_api_object, MOCK_TEST_VM
from gce_rescue.gce import Instance


class BackupTest(absltest.TestCase):
  vm : Instance


  def setUp(self):
    self.vm = Instance(test_mode=True, **MOCK_TEST_VM)
    self.vm.compute = mock_api_object(['operations'])


  def test_backup_metadata_items(self):
    """Test Backup_metadata_Items."""
    backup_data = backup.backup_metadata_items(self.vm.data)
    self.assertTrue(backup_data is not None)
    self.assertTrue('key' in backup_data[0])
    self.assertTrue('value' in backup_data[0])


  def test_backup(self):
    """Test backup task."""
    backup.backup(self.vm)


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test code for rescue.py."""

from absl.testing import absltest
from gce_rescue.gce import Instance
from gce_rescue.test.mocks import (
  mock_api_object,
  MOCK_TEST_VM,
  MOCK_TEST_DATA
)

class RescueTest(absltest.TestCase):
  vm: Instance


  def setUp(self):
    self.vm = Instance(test_mode=True, **MOCK_TEST_VM)
    self.instance_data =  self.vm.data


  def test_instance(self):
    """Compare the properties values with the fixed values in instances.json."""

    self.assertEqual(self.vm.name, 'mock_vm')
    self.assertEqual(
      self.vm.rescue_mode_status['rescue-mode'],
      False
    )
    self.assertEqual(
      self.vm.project_data,
      {
        'project': MOCK_TEST_DATA['project'],
        'zone': MOCK_TEST_DATA['zone']
      }
    )
    self.assertTrue(len(self.vm.backup_items) > 0)


  def test_define_disks(self):
    """Validate the vm._define_disks() by comparing the function returned
    dict to MOCK_TEST_DATA"""

    disks_test = self.vm._define_disks() # pylint: disable=protected-access

    self.assertEqual(
      disks_test,
      {
        'device_name': MOCK_TEST_DATA['device'],
        'disk_name': MOCK_TEST_DATA['disk']
      }
    )


  def test_refresh_fingerprint(self):
    """Validate the refresh fingerprint function by resetting the value,
    calling the function and comparing the hearded code in instances.json."""

    self.vm.compute = mock_api_object(['compute'])

    self.vm.data['metadata']['fingerprint'] = ''
    self.vm.refresh_fingerprint()

    self.assertEqual(
      self.vm.data['metadata']['fingerprint'],
      'cRcIE4_rlPM='
    )


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Standard VM operations. """

from gce_rescue.gce import Instance
from gce_rescue.tasks.keeper import wait_for_operation
import logging

_logger = logging.getLogger(__name__)

def start_instance(vm: Instance) -> str:
  """Start instance."""

  _logger.info(f'Starting {vm.name}...')
  if vm.status == 'RUNNING':
    _logger.info(f'{vm.name} is already runnning.')
    return

  operation = vm.compute.instances().start(
    **vm.project_data,
    instance = vm.name).execute()
  result = wait_for_operation(vm, oper=operation)

  if result['status'] == 'DONE':
    vm.status = 'RUNNING'
  return vm.status


def stop_instance(vm: Instance) -> str:
  """Stop instance."""
  _logger.info(f'Stopping {vm.name}...')
  if vm.status == 'TERMINATED':
    _logger.info(f'{vm.name} is already stopped.')
    return

  operation = vm.compute.instances().stop(
    **vm.project_data,
    instance = vm.name).execute()
  result = wait_for_operation(vm, oper=operation)

  if result['status'] == 'DONE':
    vm.status = 'TERMINATED'
  return vm.status




# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Different operations to guarantee VM disks backup, before performing
    any modifications."""

from gce_rescue.tasks.keeper import wait_for_operation
from typing import Dict, List
import logging

_logger = logging.getLogger(__name__)

def backup_metadata_items(data: Dict) -> List:
  """ Returns the "items" content (ssh-keys, scripts, etc) to be restored
  at the end of the process. After the instance booted and executed
  the rescue start-script
  """
  if 'items' in data['metadata'].keys():
    return data['metadata']['items']
  return []

def _create_snapshot(vm) -> Dict:
  """
  Create a snaphost of the instance boot disk, adding self._ts to the disk name.
  https://cloud.google.com/compute/docs/reference/rest/v1/disks/createSnapshot
  Returns:
    operation-result: Dict
  """

  disk_name = vm.disks['disk_name']
  snapshot_name = f'{disk_name}-{vm.ts}'
  snapshot_body = {
    'name': snapshot_name
  }
  _logger.info(f'Creating snapshot {snapshot_name}... ')
  operation = vm.compute.disks().createSnapshot(
    **vm.project_data,
    disk = disk_name,
    body = snapshot_body).execute()
  result = wait_for_operation(vm, oper=operation)
  return result

def backup(vm) -> None:
  """
  List of methods to backup data and information from the orignal instance
  """
  _create_snapshot(vm)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test code for disks.py."""

from absl.testing import absltest
from absl import logging

from gce_rescue.gce import Instance
from gce_rescue.tasks import disks
from gce_rescue.test.mocks import (
  mock_api_object,
  MOCK_TEST_VM,
  MOCK_TEST_DATA,
)

logging.set_verbosity('INFO')
# logging.set_verbosity('DEBUG')

class DisksTest(absltest.TestCase):
  def setUp(self):
    self.vm = Instance(test_mode=True, **MOCK_TEST_VM)
    self.instance_data =  self.vm.data

  def test_list_disk(self):
    self.vm.compute = mock_api_object(['disks'])
    disk_filter = f'labels.rescue={self.vm.ts}'
    result = disks.list_disk(self.vm,
      self.vm.project_data,
      disk_filter)
    self.assertTrue(len(result) > 0)
    self.assertEqual(result[0]['name'], MOCK_TEST_DATA['disk'])


  def test_attach_disk(self):
    self.vm.compute = mock_api_object([
      'disks',
      'disks',
      'operations',
      'operations',
    ])
    disk_filter = f'labels.rescue={self.vm.ts}'
    disk_list = disks.list_disk(self.vm,
      self.vm.project_data,
      disk_filter)
    disk_name = disk_list[0]['name']
    disks_ = self.instance_data['disks']
    disk = disks_[0]
    device_name = disk['deviceName']
    disks.attach_disk(
       self.vm,
       disk_name,
       device_name)


  def test_config_rescue_disks(self):
    self.vm.compute = mock_api_object([
      'operations',
      'operations',
      'operations',
      'operations',
      'operations',
      'disks',
    ])
    disks.config_rescue_disks(self.vm)


  def test_restore_original_disk(self):
    self.vm.compute = mock_api_object([
      'operations',
      'operations',
      'operations',
      'operations',
    ])
    disks.restore_original_disk(self.vm)


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test code for metadata.py."""

from absl.testing import absltest

from gce_rescue.gce import Instance
from gce_rescue.tasks import metadata
from gce_rescue.test.mocks import  mock_api_object, MOCK_TEST_VM


class MetadataTest(absltest.TestCase):
  def setUp(self):
    self.vm = Instance(test_mode=True, **MOCK_TEST_VM)
    self.instance_data =  self.vm.data


  def test_set_metadata(self):
    self.vm.compute = mock_api_object(['operations'])
    result = metadata.set_metadata(self.vm)
    self.assertTrue(len(result)> 1)


  def test_restore_metadata(self):
    self.vm.ts = 1666774335
    self.vm.compute = mock_api_object([
      'compute',
      'serialconsole',
      'operations',
    ])
    result = metadata.restore_metadata_items(self.vm)
    self.assertTrue(len(result)> 1)

if __name__ == '__main__':
  absltest.main()




# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=broad-exception-raised

"""keeper that the progress of the tasks. """

import googleapiclient.discovery
from time import sleep
from typing import Dict
import logging
import json

_logger = logging.getLogger(__name__)


def wait_for_operation(
  instance_obj: googleapiclient.discovery.Resource,
  oper: Dict
) -> Dict:
  """ Creating poll to wait the operation to finish. """

  while True:
    if oper['status'] == 'DONE':
      _logger.info('done.')
      if 'error' in oper:
        raise Exception(oper['error'])
      return oper

    oper = instance_obj.compute.zoneOperations().get(
      **instance_obj.project_data,
      operation = oper['name']).execute()
    sleep(1)

def wait_for_os_boot(vm: googleapiclient.discovery.Resource) -> bool:
  """Wait guest OS to complete the boot proccess."""

  timeout = 60
  wait_time = 2
  end_string = f'END:{vm.ts}'
  _logger.info('Waiting startup-script to complete.')
  while True:
    result = vm.compute.instances().getSerialPortOutput(
      **vm.project_data,
      instance = vm.name
    ).execute()

    if end_string in json.dumps(result):
      _logger.info('startup-script has ended.')
      return True

    sleep(wait_time)
    timeout -= wait_time
    if not timeout:
      return False


# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Compilations of all disks tasks related. """

from typing import Dict
import logging

import googleapiclient.errors

from gce_rescue.tasks.keeper import wait_for_operation
from gce_rescue.tasks.backup import  backup
from gce_rescue.utils import ThreadHandler as Handler

_logger = logging.getLogger(__name__)

def _create_rescue_disk(vm, source_disk: str) -> Dict:
  """ Create new temporary rescue disk based on source_disk.
  https://cloud.google.com/compute/docs/reference/rest/v1/disks/insert
  Returns:
    operation-result: Dict
  """

  chk_disk_exist = {}
  try:
    chk_disk_exist = vm.compute.disks().get(
      **vm.project_data,
      disk = vm.rescue_disk).execute()
  except googleapiclient.errors.HttpError as e:
    if e.status_code == 404:
      _logger.info(f'Creating rescue disk {vm.rescue_disk}...')
    else:
      raise Exception(e) from e

  if 'name' in chk_disk_exist.keys():
    if 'users' in chk_disk_exist.keys():
      disk_users = chk_disk_exist['users']
      raise Exception(
        f'Disk {vm.rescue_disk} is currently in use: {disk_users}'
      )
    _logger.info(f'Disk {vm.rescue_disk} already exist. Skipping...')
    return {}

  disk_body = {
    'name': vm.rescue_disk,
    'sourceImage': source_disk,
    'type': f'projects/{vm.project}/zones/{vm.zone}/diskTypes/pd-balanced'
  }
  operation = vm.compute.disks().insert(
    **vm.project_data,
    body = disk_body).execute()

  result = wait_for_operation(vm, oper=operation)
  return result


def _set_disk_label(vm, disk_name = str) -> Dict:
  """ Set labels.rescue=TS to be able to idenfied the boot disk when restore
  the VM to the normal configuration.
  https://cloud.google.com/compute/docs/reference/rest/v1/disks/setLabels
  Return:
    operation-result: Dict
  """

  name_filter = f'name={disk_name}'
  response = list_disk(
    vm,
    project_data=vm.project_data,
    label_filter=name_filter
  )
  label_fingerprint = response[0]['labelFingerprint']
  request_body = {
    'labels': {
        'rescue': vm.ts
    },
    'labelFingerprint': label_fingerprint
  }
  operation = vm.compute.disks().setLabels(
    **vm.project_data,
    resource = disk_name,
    body = request_body).execute()

  return operation

def _delete_rescue_disk(vm, disk_name: str) -> Dict:
  """
  Delete rescue disk after resetting the instance to the original configuration.
  https://cloud.google.com/compute/docs/reference/rest/v1/disks/delete
  Param:
    disk_name: str, Name of the disk to be deleted.
  Returns:
    operation-result: Dict
  """
  _logger.info(f'Deleting disk {disk_name}...')
  operation = vm.compute.disks().delete(
    **vm.project_data,
    disk = disk_name).execute()

  result = wait_for_operation(vm, oper=operation)
  return result


def list_disk(vm, project_data: Dict, label_filter: str) -> Dict:
  """Filter existing disks with labels.rescue=TS
  https://cloud.google.com/compute/docs/reference/rest/v1/disks/list
  Returns:
    result: Dict
  """
  result = vm.compute.disks().list(
    **project_data,
    filter=label_filter).execute()

  #TODO: Add validation and throw exception if response has more than 1 disk:
  #  len(response['items'])
  #For test phase - remember to fix that
  return result['items']


def attach_disk(
  vm,
  disk_name: str,
  device_name: str,
  boot: bool = False
) -> Dict:
  """
  Attach disk on the instance. By default it will attach as secundary
  https://cloud.google.com/compute/docs/reference/rest/v1/instances/attachDisk
  Returns:
      operation-result: Dict
  """
  if not boot:
    request = _set_disk_label(vm, disk_name)
    if request['status'] != 'DONE':
      _logger.error(f'Unable to set label to disk {disk_name}.')
      raise Exception(request)
    else:
      _logger.info(f'Label configured successfully disk {disk_name}.')
  attach_disk_body = {
    'boot': boot,
    'name': disk_name,
    'deviceName': device_name,
    'type': 'PERSISTENT',
    'source': f'projects/{vm.project}/zones/{vm.zone}/disks/{disk_name}'
  }
  _logger.info(f'Attaching disk {disk_name}...')
  operation = vm.compute.instances().attachDisk(
    **vm.project_data,
    instance = vm.name,
    body = attach_disk_body).execute()

  result = wait_for_operation(vm, oper=operation)
  return result


def _detach_disk(vm, disk: str) -> Dict:
  """ Detach disk from the instance.
  https://cloud.google.com/compute/docs/reference/rest/v1/instances/detachDisk
  """

  _logger.info(f'Detaching disk {disk} from {vm.name}...')
  operation = vm.compute.instances().detachDisk(
      **vm.project_data,
      instance = vm.name,
      deviceName = disk).execute()
  result = wait_for_operation(vm, oper=operation)
  return result


def config_rescue_disks(vm) -> None:
  device_name = vm.disks['device_name']
  backup(vm)
  # task1 = multitasks.Handler(
  #     target = backup,
  #     kwargs={'vm' : vm}
  #     )
  # task1.start()
  task2 = Handler(
    target = _create_rescue_disk,
    kwargs={'vm': vm, 'source_disk': vm.rescue_source_disk}
    )
  task2.start()
  task2.join()
  _detach_disk(vm, disk=device_name)
  attach_disk(
    vm,
    disk_name=vm.rescue_disk,
    device_name=vm.rescue_disk,
    boot=True
  )

def restore_original_disk(vm) -> None:
  """ Restore tasks to the original disk """
  device_name = vm.disks['device_name']

  _detach_disk(vm, disk=vm.rescue_disk)
  _detach_disk(vm, disk=device_name)
  _delete_rescue_disk(vm, disk_name=vm.rescue_disk)
  attach_disk(vm, **vm.disks, boot=True)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Mock API for tests purposes."""

import os
import pathlib
import json
from typing import List
import googleapiclient.discovery
from googleapiclient.http import HttpMockSequence


MOCK_TEST_VM = {
  'project': 'mock_project',
  'zone': 'europe-central1-a',
  'name' : 'mock_vm'
}

MOCK_TEST_DATA = {
  'project': 'mock_project',
  'zone': 'europe-central1-a',
  'name' : 'mock_vm',
  'disk' : 'mock-vm',
  'device': 'persistent-disk-0'
}

TESTDATA_PATH='test-data'
mock_data = {
  'compute': f'{TESTDATA_PATH}/instances.json',
  'disks': f'{TESTDATA_PATH}/disks.json',
  'operations': f'{TESTDATA_PATH}/operations.json',
  'serialconsole': f'{TESTDATA_PATH}/serialconsole.json',
}


def mock_api_object(mocks_list: List[str]):
  """ Returns mock HTTP sequence Resources to be used in API tests calls """
  responses = []
  for mock in mocks_list:
    if mock not in mock_data:
      raise Exception(ValueError, mock)
    file_name = os.path.join(os.path.dirname(__file__), mock_data[mock])
    if not pathlib.Path(file_name).is_file():
      raise Exception(FileNotFoundError, file_name)
    with open(file_name, encoding='utf-8') as fd:
      data = json.load(fd)
    responses.append(
      ({'status': '200'}, json.dumps(data))
    )
  http = HttpMockSequence(responses)
  service = googleapiclient.discovery.build('compute', 'v1', http = http)
  return service



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Authentication validation to be called from ../pre_validations.py """
import google.auth
import sys

from googleapiclient.discovery import Resource
from gce_rescue.tasks.validations.api import api_service
from gce_rescue.test.mocks import mock_api_object

PROJECT = ''

def _get_auth():
  global PROJECT
  try:
    credentials, adc_project = google.auth.default()
    if not adc_project and not PROJECT:
      msg = _info_no_project()
      print(msg, file=sys.stderr)
      sys.exit(1)
    if not PROJECT and adc_project:
      PROJECT = adc_project
    return credentials
  except google.auth.exceptions.DefaultCredentialsError:
    msg = _info_auth_cred()
    print(msg, file=sys.stderr)
    sys.exit(1)

def authenticate_check(
  zone: str,
  instance_name: str,
  project: str = None,
  test_mode: bool = False
) -> Resource:
  global PROJECT
  PROJECT = project
  if test_mode:
    service = mock_api_object(['compute'])
    return service
  credentials = _get_auth()
  if not credentials:
    return False
  # service = googleapiclient.discovery.build(
  #   'compute',
  #   'v1',
  #   credentials = credentials
  # )
  service =  api_service('compute', 'v1', credentials)
  request = service.instances().get(
		project = PROJECT,
		zone = zone,
		instance = instance_name)
  try:
    request.execute()
    return service
  except google.auth.exceptions.RefreshError:
    msg = _info_auth_refresh()
    print(msg, file=sys.stderr)
    sys.exit(1)

def project_name() -> str:
  return PROJECT

def _info_auth_refresh() -> str:
  return (
	'    Please use application-default Credentials (ADC) to authenticate:\n'
	'    $ gcloud auth login --update-adc'
  )

def _info_auth_cred() -> str:
  return (
	'    Please use application-default Credentions (ADC) to authenticate:\n'
	'     $ gcloud auth application-default login\n'
	f'     $ gcloud auth application-default set-quota-project {PROJECT}'
  )

def _info_no_project() -> str:
  return (
	'    Was not possible to find the project where the VM is created.\n'
	'    You can use the option --project to declare the project-id or '
  'set to your configuration:\n'
	'    $ gcloud config set project PROJECT_ID'
  )



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Authorization validation to be called from ../pre_validations.py
Permissions:
  compute.instances.stop
  compute.instances.start
  compute.instances.attachDisk
  compute.instances.detachDisk
  compute.images.useReadOnly
  compute.disks.use
  compute.disks.setLabels
  compute.snapshots.create
  compute.disks.createSnapshot
  compute.instances.setMetadata
  compute.instances.setLabels
"""
import google.auth
from gce_rescue.tasks.validations.api import api_service

def authorize_check(project: str = None) -> bool:

  permissions_list = ['compute.snapshots.create']
  body_data = {'permissions': permissions_list}
  credentials, project_id = google.auth.default()

  if not project:
    project = project_id

  service = api_service('cloudresourcemanager', 'v1', credentials)
  result = service.projects().testIamPermissions(
    resource = project,
    body =  body_data
  ).execute()

  if permissions_list != result['permissions']:
    raise PermissionError()

  return True



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Validations that should be performed before continue the script execution.
    Each validations steps can be appended to the class Validations() and keep
    the file under validations/ folder."""

from dataclasses import dataclass, field
import googleapiclient.discovery
from gce_rescue.tasks.validations.authorization import authorize_check
from gce_rescue.tasks.validations.authentication import (
  authenticate_check,
  project_name
)

@dataclass
class Validations:
  """Run different validations before continue the script.
  	Return:
  		discovery.build object.
  """
  zone: str
  name: str
  project: str = None
  test_mode: bool = field(default_factory=False)

  def _authentication(self):
    return authenticate_check(
      project = self.project,
      zone = self.zone,
      instance_name = self.name,
      test_mode = self.test_mode,
    )

  def __post_init__(self):
    if not self.test_mode:
      authorize_check(project = self.project)

  @property
  def compute(self) -> googleapiclient.discovery.Resource:
    return self._authentication()

  @property
  def adc_project(self) -> str:
    return project_name()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Common API objects """
import googleapiclient
import google_auth_httplib2
import httplib2

from google.oauth2.credentials import Credentials
from googleapiclient.discovery import Resource

from gce_rescue.config import VERSION

def api_service(
    service: str,
    version: str,
    credentials: Credentials) -> Resource:

  def _builder(http, *args, **kwargs):
    # google api client is not thread safe
    # https://github.com/googleapis/google-api-python-client/blob/main/docs/thread_safety.md
    del http
    headers = kwargs.setdefault('headers',{})
    headers['user-agent'] = f'gce_rescue-{VERSION}'
    auth_http = google_auth_httplib2.AuthorizedHttp(credentials,
                                                   http=httplib2.Http())
    return googleapiclient.http.HttpRequest(auth_http, *args, **kwargs)

  service_ = googleapiclient.discovery.build(service, version,
                        cache_discovery=False,
                        credentials=credentials,
                        requestBuilder=_builder)
  return service_



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" List of ordered tasks to be executed when set/reset VM rescue mode. """

from typing import List
import logging

from gce_rescue.gce import Instance
from gce_rescue.tasks.disks import (
  config_rescue_disks,
  restore_original_disk,
  attach_disk
)
from gce_rescue.tasks.operations import (
  start_instance,
  stop_instance
)
from gce_rescue.tasks.metadata import (
  set_metadata,
  restore_metadata_items
)
from gce_rescue.utils import Tracker

_logger = logging.getLogger(__name__)

def _list_tasks(vm: Instance, action: str) -> List:
  """ List tasks, by order, per operation
    operations (str):
      1. set_rescue_mode
      2. reset_rescue_mode
  """
  all_tasks = {
    'set_rescue_mode': [
      {
        'name': stop_instance,
        'args': [{
          'vm': vm
        }]
      },
      {
        'name': config_rescue_disks,
        'args': [{
          'vm': vm
        }]
      },
      {
        'name': set_metadata,
        'args': [{
          'vm': vm
        }]
      },
      {
        'name': start_instance,
        'args': [{
          'vm': vm
        }]
      },
      {
        'name': attach_disk,
        'args': [{
          'vm': vm,
          'boot': False,
          **vm.disks
        }],
      },
      {
        'name': restore_metadata_items,
        'args': [{
          'vm': vm
        }],
      }
  ],
    'reset_rescue_mode': [
      {
        'name': stop_instance,
        'args': [{
          'vm': vm
        }]
      },
      {
        'name': restore_original_disk,
        'args': [{
          'vm': vm
        }]
      },
      {
        'name': restore_metadata_items,
        'args': [{
          'vm': vm,
          'remove_rescue_mode': True
        }]
      },
      {
        'name': start_instance,
        'args': [{
          'vm': vm
        }]
      },
    ]
  }

  if action not in all_tasks:
    _logger.info(f'Unable to find "{action}".')
    raise ValueError()
  return all_tasks[action]


def call_tasks(vm: Instance, action: str) -> None:
  """ Loop tasks dict and execute """
  tasks = _list_tasks(vm = vm, action = action)
  total_tasks = len(tasks)

  tracker = Tracker(total_tasks)
  tracker.start()

  for task in tasks:
    execute = task['name']
    args = task['args'][0]

    execute(**args)
    tracker.advance(step = 1)

  tracker.finish()


# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" List of messages to inform and educate the user. """

from gce_rescue.gce import Instance

def tip_connect_ssh(vm: Instance) -> str:
  return (f'└── Your instance is READY! You can now connect your instance '
    f' {vm.name} via:\n  1. CLI. (add --tunnel-through-iap if necessary)\n'
    f'    $ gcloud compute ssh {vm.name} --zone={vm.zone} '
    f'--project={vm.project} --ssh-flag="-o StrictHostKeyChecking=no"\n  OR\n'
    f'  2. Google Cloud Console:\n'
    f'    https://ssh.cloud.google.com/v2/ssh/projects/{vm.project}/zones/'
    f'{vm.zone}/instances/{vm.name}?authuser=0&hl=en_US&useAdminProxy=true&'
    f'troubleshoot4005Enabled=true\n')

def tip_restore_disk(vm: Instance) -> str:
  return (f'└── The instance {vm.name} was restored! Use the snapshot below '
    f'if you need to restore the modification made while the instance was '
    f'in rescue mode.\n Snapshot name: {vm.disks["disk_name"]}-{vm.ts}\n'
    f' More information: '
    f'https://cloud.google.com/compute/docs/disks/restore-snapshot\n')



#!/usr/bin/env python3

# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Install GCE Rescue """

from setuptools import setup, find_packages
from os import path

from gce_rescue.config import VERSION

my_pwd = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(my_pwd, 'README.md'), encoding='utf-8') as f:
  long_description_readme = f.read()

# Get requirements
with open(path.join(my_pwd, 'requirements.txt'), encoding='utf-8') as f:
  list_requirements = f.readlines()

setup(
  name = 'gce-rescue',
  version = VERSION,
  description='GCE Rescue - Boot your GCE VM in rescue mode.',
  url = 'https://github.com/googlecloudplatform/gce-rescue',
  author = 'Halley de Souza',
  author_email = 'gce-rescue-dev@google.com',
  license = 'Apache-2.0',
  long_description = long_description_readme,
  long_description_content_type = 'text/markdown',
  install_requires = list_requirements,
  zip_safe = False,
  packages = find_packages(exclude=['test', '*_test.py']),
  package_dir = {'': '.'},
  package_data = {'': ['startup-script.txt']},
  include_package_data = True,
  entry_points={
        'console_scripts': [
            'gce-rescue = gce_rescue.bin.rescue:main',
        ],
    },
  classifiers = [
    'License :: OSI Approved :: Apache Software License',
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Programming Language :: Python :: 3.11',
    'Programming Language :: Python :: 3.12',
    'Topic :: System :: Boot',
    'Topic :: System :: Boot :: Init',
    'Topic :: System :: Recovery Tools',
    'Topic :: System :: System Shells',
    'Topic :: System :: Systems Administration',
  ]
)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Default configurations values. """

import os
import argparse

dirname = os.path.dirname(__file__)

VERSION = '0.3'

config = {
  'version': VERSION,
  'debug': False,
  'startup-script-file': os.path.join(dirname, 'startup-script.txt'),
  'source_guests': {
    'x86_64':[
      'projects/debian-cloud/global/images/family/debian-11',
      'projects/rocky-linux-cloud/global/images/family/rocky-linux-9'
      ],
    'arm64':[
      'projects/debian-cloud/global/images/family/debian-11-arm64',
      'projects/rocky-linux-cloud/global/images/family/rocky-linux-9-arm64'
    ]
  }
}

def get_config(key):
  if key in config:
    return config[key]


def process_args():
  """ Print usage options. """
  parser = argparse.ArgumentParser(description=f'GCE Rescue v{VERSION} - \
    Set/Reset GCE instances to boot in rescue mode.')
  parser.add_argument('-p', '--project',
                      help='The project-id that has the instance.')
  parser.add_argument('-z', '--zone', help='Zone where the instance \
    is created.',
                      required=True)
  parser.add_argument('-n', '--name', help='Instance name.', required=True)
  parser.add_argument('-d', '--debug', action='store_true',
                      help='Print to the log file in debug leve')
  parser.add_argument('-f', '--force', action='store_true',
                      help='Don\'t ask for confirmation.')

  return parser


def set_configs(user_args):
  config['debug'] = getattr(user_args, 'debug')



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=broad-exception-raised

""" List of classes and functions to be used across the code. """

from time import sleep
import logging
import multiprocessing
from threading import Thread
import sys
from gce_rescue.config import get_config


_logger = logging.getLogger(__name__)

class Tracker():
  """ Track tasks using multiprocessing and print progress bar. """

  def __init__(self, target):
    self.target = target
    self._pivot = multiprocessing.Value('i', 1)
    self._proc = None

  def start(self):
    self._proc = multiprocessing.Process(target=self._run)
    self._proc.start()
    print('┌── Configuring...')

  def advance(self, step=None):
    if not step:
      step = 1
    self._pivot.value += step

  def finish(self):
    self._pivot.value = self.target
    self._proc.join()
    sleep(0.5)
    print('├── Configurations finished.')

  def _loading(self):
    chars = ['-', '|', '/', '|', '\\']
    i = 0
    while True:
      yield chars[i]
      i += 1
      if i == len(chars):
        i = 0

  def _run(self):
    self._gen = self._loading()
    while self._pivot.value < self.target:
      try:
        sleep(0.001)
        self._print()
      except Exception as exc:
        raise f'{exc}: {self._pivot.value} = {self.target}'

    self._print()
    print('\r')

  def _print(self):
    size = 60
    loading = next(self._gen)
    if self._pivot.value == self.target:
      loading = '█'
    count = self._pivot.value
    total = self.target
    x = int(size * self._pivot.value / self.target)
    progress = '█' * x
    bar = '.' * (size-x)
    print(f'│   └── Progress {count}/{total} [{progress}{loading}{bar}]',
      end='\r',
      file=sys.stderr,
      flush=True)


class ThreadHandler(Thread):
  """Handler for multithread tasks."""

  def __init__(
      self,
      group=None,
      target=None,
      name=None,
      args=None,
      kwargs=None
    ):

    if not args:
      args = ()

    if not kwargs:
      kwargs = {}

    Thread.__init__(self, group, target, name, args, kwargs)
    self._result = None

  def run(self):
    if self._target is not None:
      self._result = self._target(*self._args, **self._kwargs)

  def result(self, *args):
    Thread.join(self, *args)
    return self._result


def set_logging(vm_name: str) -> None:
  """ Set logfile and verbosity. """

  level = 'DEBUG' if get_config('debug') else 'INFO'
  log_level = getattr(logging, level.upper())
  file_name = f'{vm_name}.log'
  logging.basicConfig(
    filename=file_name,
    filemode='a',
    format='%(asctime)s,%(msecs)03d %(levelname)-8s [%(filename)s:%(lineno)d]\
      %(message)s',
    datefmt='%Y-%m-%d:%H:%M:%S',
    level=log_level)


def read_input(msg: str) -> None:
  """Read user input if --force is not provided."""
  print(msg, end='')
  input_answer = input()
  input_answer = input_answer.strip()
  if input_answer.upper() != 'Y':
    print(f'got input: "{input_answer}". Aborting')
    sys.exit(1)
    